import argparse
import asyncio
import importlib
import uuid

import aiohttp
import aiohttp.web as aiohttp_web


class MultiClientServer:
    def __init__(self, trame_app, max_msg_size=10000000, heartbeat=30):
        self._completion = None
        self.trame_app = trame_app
        self.ws_options = dict(max_msg_size=max_msg_size, heartbeat=heartbeat)
        self.web_server = aiohttp_web.Application()
        self.web_server.router.add_route("GET", "/", self._index_handler)

        routes = [aiohttp_web.get("/ws", self._ws_handler)]
        app_www = self.trame_app("__www__").server
        for route in sorted(app_www.serve.keys(), reverse=True):
            routes.append(
                aiohttp_web.static(
                    f"/{route}", app_www.serve[route], append_version=True
                )
            )
        routes.append(aiohttp_web.static("/", app_www._www, append_version=True))
        self.web_server.add_routes(routes)

    async def _index_handler(self, request):
        if request.query_string:
            return aiohttp.web.HTTPFound(f"index.html?{request.query_string}")
        return aiohttp.web.HTTPFound("index.html")

    async def _ws_handler(self, request):
        name = str(uuid.uuid4()).replace("-", "")

        # App setup
        print(f" + {name}")
        app = self.trame_app(name)
        task = app.server.start(backend="generic", exec_mode="task")
        await app.server.ready

        ws_network = aiohttp_web.WebSocketResponse(**self.ws_options)

        async def on_msg_from_server(binary, content):
            if binary:
                await ws_network.send_bytes(content)
            else:
                await ws_network.send_str(content)

        try:
            await ws_network.prepare(request)
            ws_app = app.server._server.ws
            connection = ws_app.connect()
            connection.on_message(on_msg_from_server)
            async for msg in ws_network:
                await connection.send(msg.type == aiohttp.WSMsgType.BINARY, msg)
        finally:
            connection.close()
            await app.server.stop()
            await task

        print(f" - {name}")
        return ws_network

    async def run(self, host="localhost", port=8080):
        self._completion = asyncio.get_event_loop().create_future()
        runner = aiohttp_web.AppRunner(
            self.web_server,
            handle_signals=True,
        )
        await runner.setup()

        site = aiohttp_web.TCPSite(runner, host, port)

        await site.start()
        await self._completion

    def stop(self):
        if self._completion is not None:
            self._completion.set_result(True)


def main():
    parser = argparse.ArgumentParser(description="Serve trame application")

    parser.add_argument(
        "--exec",
        default="trame.app.demo:Cone",
        help="Trame app to serve (default: trame.app.demo:Cone)",
    )
    parser.add_argument(
        "--host",
        default="localhost",
        help="IP or hostname to serve on (default: localhost)",
    )
    parser.add_argument(
        "--port",
        default=8080,
        type=int,
        help="Port to serve on (default: 8080)",
    )

    parser.add_argument(
        "--ws-heart-beat",
        default=30,
        type=int,
        help="WebSocket heart beat (default: 30)",
    )

    parser.add_argument(
        "--ws-max-size",
        default=10000000,
        type=int,
        help="WebSocket maximum message size (default: 10000000)",
    )

    args, _ = parser.parse_known_args()
    if args.exec is None or ":" not in args.exec:
        parser.print_help()
        return

    module_path, app_name = args.exec.split(":")
    try:
        module = importlib.import_module(module_path)
    except ModuleNotFoundError:
        print(f"Invalid trame app to serve: {args.app}")
        parser.print_help()
        return

    print("Server configuration")
    print(f"  - host: {args.host}")
    print(f"  - port: {args.port}")
    print(f"  - app: {app_name} from {module_path}")
    print("Websocket configuration")
    print(f"  - heartbeat: {args.ws_heart_beat}")
    print(f"  - Max message size: {args.ws_max_size}")

    trame_app = getattr(module, app_name)
    web_server = MultiClientServer(
        trame_app, heartbeat=args.ws_heart_beat, max_msg_size=args.ws_max_size
    )
    asyncio.run(web_server.run(args.host, args.port))


if __name__ == "__main__":
    main()



import logging
from trame_server import Server, Client
from trame_server.core import set_default_client_type
from trame_client.widgets.core import VirtualNode

# Ensure this is imported so that mimetypes.init() is decorated
import trame.app.mimetypes  # noqa: F401

DEFAULT_NAME = "trame"
AVAILABLE_SERVERS = {}
AVAILABLE_CLIENTS = {}

logger = logging.getLogger(__name__)

set_default_client_type("vue3")


def apply_client_type(server, client_type=None):
    if client_type is not None:
        server.client_type = client_type
    return server


def get_server(name=None, create_if_missing=True, client_type=None, **kwargs):
    """
    Return a server for serving trame applications.
    If a name is given and such server is not available yet,
    it will be created otherwise the previously created instance will be returned.

    :param name: A server name identifier which can be useful when several servers
                are expected to be created. Most of the time, passing no arguments
                is what you are looking for. Also an actual Server instance can be
                provided so you can use it as a decorator function.
    :type name: None | str | Server instance

    :param create_if_missing: By default if a server for a given name does not exist
        that method will create it.
    :type create_if_missing: bool

    :param client_type: If provided, it will set it on the server.
    :type client_type: None | "vue2" | "vue3"

    :param **kwargs: any extra keyword args are passed as option to the server instance.

    :return: Return a unique Server instance per given name.
    :rtype: trame_server.core.Server
    """
    # Convenient method for decorator like usage
    if isinstance(name, Server):
        return apply_client_type(name, client_type)

    if name is None:
        name = DEFAULT_NAME

    if name in AVAILABLE_SERVERS:
        return apply_client_type(AVAILABLE_SERVERS[name], client_type)

    if create_if_missing:
        server = Server(name, VirtualNode, **kwargs)
        AVAILABLE_SERVERS[name] = server

        return apply_client_type(server, client_type)

    # No server available for given name
    return None


def get_client(url=None, hot_reload=False, **kwargs):
    """
    Return a client to a remote trame applications.
    If a url is given and such client is not available yet,
    it will be created otherwise the previously created instance will be returned.

    :param url: Websocket URL which to connect to.
    :type url: str


    :param hot_reload: Enable when state change function should be hot reloaded.
    :type hot_reload: bool

    :param **kwargs: any extra keyword args use for authentication configuration.

    :return: Return a unique Client instance per given url. Each instance need to a connect() call.
    :rtype: trame_server.client.Client
    """
    if url in AVAILABLE_CLIENTS:
        return AVAILABLE_CLIENTS[url]

    client = Client(url=url, config=kwargs, hot_reload=hot_reload)
    if url is not None:
        AVAILABLE_CLIENTS[url] = client

    return client


__all__ = [
    "get_server",
    "get_client",
]



import re

camel_pattern = re.compile(r"(?<!^)(?=[A-Z])")


def camel_to_dash(value):
    "MySuperWidget => my-super-widget"
    return camel_pattern.sub("-", value).lower()


def attr_to_py(value):
    """
    top-left                => "top_left"
    top-left:default        => ("top_left_default", "top-left:default")
    v-slot:default.modifier => ("v_slot_default_modifier", "v-slot:default.modifier")
    """
    py_name = value.replace("-", "_")
    if "." in value or ":" in value:
        py_name = py_name.replace(".", "_").replace(":", "_")
        return [py_name, value]
    return py_name



import yaml
import shutil
from pathlib import Path
import aiohttp

PKG_INIT = """__path__ = __import__("pkgutil").extend_path(__path__, __name__)"""

EXT_BY_TYPES = {
    "scripts": ".js",
    "styles": ".css",
}


async def generate_trame_package(config_path, output_path):
    output = Path(output_path)
    output.mkdir(parents=True, exist_ok=True)
    with open(config_path) as f:
        config = yaml.safe_load(f)

    await create_base_structure(Path(config_path).parent.absolute(), config, output)


async def create_base_structure(ref_path, config, output):
    trame_plugins = {}

    # Trame structure
    for package_name in ["modules", "widgets"]:
        root_dir = output / "trame" / package_name
        root_dir.mkdir(parents=True, exist_ok=True)
        (root_dir / "__init__.py").write_text(PKG_INIT)
    (output / "trame" / "__init__.py").write_text(PKG_INIT)

    # Package structure
    for name in config:
        root_dir = output / name
        root_dir.mkdir(parents=True, exist_ok=True)
        (root_dir / "__init__.py").write_text("")
        for sub_name in config[name]:
            sub_package = root_dir / sub_name
            sub_package.mkdir(parents=True, exist_ok=True)
            (sub_package / "__init__.py").write_text("")

            # Create files
            for module in config[name][sub_name]:
                if sub_name == "module":
                    module_root = sub_package / module
                    module_root.mkdir(parents=True, exist_ok=True)
                    module_root_init = module_root / "__init__.py"
                    module_conf_init = {}
                    for web_dir in config[name][sub_name][module]:
                        module_conf_init[web_dir] = await create_web_content(
                            ref_path,
                            (module_root / web_dir),
                            config[name][sub_name][module][web_dir],
                        )
                    # Register trame/modules/{webdir}
                    trame_plugins[
                        f"trame/modules/{module}.py"
                    ] = f"from {name}.module.{module} import *\n"
                    create_module_init(module_root_init, module_conf_init)

                elif sub_name == "widgets":
                    file_path = sub_package / f"{module}.py"
                    with file_path.open("w") as file:
                        file.write(
                            "from trame_client.widgets.core import AbstractElement, Template  # noqa\n"
                        )
                        file.write(f"from ..module import {module}\n")
                        file.write("\nclass HtmlElement(AbstractElement):")
                        file.write(
                            "\n    def __init__(self, _elem_name, children=None, **kwargs):"
                        )
                        file.write(
                            "\n        super().__init__(_elem_name, children, **kwargs)"
                        )
                        file.write("\n        if self.server:")
                        file.write(f"\n            self.server.enable_module({module})")
                        file.write("\n\n")

                        all_class_names = []
                        for class_name, class_info in config[name][sub_name][
                            module
                        ].items():
                            if class_name == "directives":
                                for entry in class_info:
                                    if isinstance(entry, list):
                                        file.write(
                                            f'\nAbstractElement.register_directive("{entry[0]}", "{entry[1]}")'
                                        )
                                    else:
                                        file.write(
                                            f'\nAbstractElement.register_directive("{entry}")'
                                        )
                                file.write("\n\n")
                            else:
                                all_class_names.append(class_name)
                                component = class_info.get("component")
                                properties = class_info.get("properties", [])
                                events = class_info.get("events", [])
                                file.write(f"\nclass {class_name}(HtmlElement):")
                                to_py_help(file, class_info)
                                if class_info.get("methods"):
                                    file.write("\n    _next_id = 0")
                                file.write(
                                    "\n    def __init__(self, children=None, **kwargs):"
                                )
                                file.write(
                                    f'\n        super().__init__("{component}", children, **kwargs)'
                                )
                                file.write("\n        self._attr_names += [")
                                for item in properties:
                                    to_py_attr(file, item, 4 * 3)
                                file.write("\n        ]")
                                file.write("\n        self._event_names += [")
                                for item in events:
                                    to_py_attr(file, item, 4 * 3)
                                file.write("\n        ]")

                                # Generate ref
                                if class_info.get("methods"):
                                    file.write(
                                        f"""\n        {class_name}._next_id += 1"""
                                    )
                                    file.write(
                                        f"""\n        self.__ref = kwargs.get("ref", f"{class_name}_{{{class_name}._next_id}}")"""
                                    )
                                    file.write(
                                        """\n        self._attributes["ref"] = f'ref="{self.__ref}"'"""
                                    )
                                    file.write("\n\n")

                                    file.write("\n    @property")
                                    file.write("\n    def ref(self):")
                                    file.write("\n        return self.__ref")
                                    file.write("\n\n")

                                    for entry in class_info.get("methods"):
                                        method = entry.get("name")
                                        help = entry.get("help")
                                        if isinstance(method, (list, tuple)):
                                            py_m, js_m = method
                                        else:
                                            py_m = method
                                            js_m = method
                                        file.write(f"\n    def {py_m}(self, *args):")
                                        if help:
                                            file.write(f'\n        """{help}')
                                            if help.endswith("\n"):
                                                file.write('        """')
                                            elif "\n" in help:
                                                file.write('\n        """')
                                            else:
                                                file.write('"""')
                                        file.write(
                                            f'\n        self.server.js_call(self.ref, "{js_m}", *args)'
                                        )
                                        file.write("\n\n")
                                else:
                                    file.write("\n\n")

                        file.write("\n__all__ = [")
                        for class_name in all_class_names:
                            file.write(f'\n{" " * 4}"{class_name}",')
                        file.write("\n]")
                        file.write("\n")

                    # Register trame/modules/{webdir}
                    trame_plugins[
                        f"trame/widgets/{module}.py"
                    ] = f"from {name}.widgets.{module} import *\n\ndef initialize(server):\n    from {name}.module import {module}\n\n    server.enable_module({module})\n"

    # Create trame package connectors
    for file, content in trame_plugins.items():
        (output / file).write_text(content)


def create_module_init(init_path, init_conf):
    for name in init_conf:
        m_conf = init_conf[name]
        m = init_path.with_name(f"{name}.py")
        widget_name = init_path.parent.name
        with m.open("w") as file:
            file.write("from pathlib import Path\n")
            file.write("\n")
            file.write(
                f'serve_path = str(Path(__file__).with_name("{name}").resolve())\n'
            )
            file.write(f'serve = {{"__trame_{widget_name}": serve_path }}\n')
            for group in ["scripts", "styles"]:
                if group in m_conf:
                    file.write(f"{group} = [\n")
                    for f_name in m_conf[group]:
                        file.write(f'    "__trame_{widget_name}/{f_name}",\n')
                    file.write("]\n")
            if "vue_use" in m_conf:
                file.write("vue_use = [\n")
                for name in m_conf["vue_use"]:
                    file.write(f'    "{name}",\n')
                file.write("]\n")

    with init_path.open("w") as file:
        file.write("def setup(server, **kargs):\n")
        file.write('    client_type = "vue3"\n')
        file.write('    if hasattr(server, "client_type"):\n')
        file.write("        client_type = server.client_type\n")
        if "vue2" in init_conf:
            file.write('    if client_type == "vue2":\n')
            file.write("        from . import vue2\n")
            file.write("        server.enable_module(vue2)\n")

        if "vue3" in init_conf:
            file.write(
                f'    {"elif" if "vue2" in init_conf else "if" } client_type == "vue3":\n'
            )
            file.write("        from . import vue3\n")
            file.write("        server.enable_module(vue3)\n")
        file.write("    else:\n")
        file.write("        raise TypeError(\n")
        file.write(
            '            f"Trying to initialize trame_vuetify with unknown client_type={client_type}"\n'
        )
        file.write("        )\n")


async def create_web_content(ref_path, base_directory, web_config):
    out_conf = {}
    base_directory.mkdir(parents=True, exist_ok=True)
    for key in web_config:
        if key in ["scripts", "styles"]:
            local_conf = []
            out_conf[key] = local_conf
            for item in web_config[key]:
                if isinstance(item, str):
                    if item.startswith("http"):
                        local_conf.append(
                            await handle_url(base_directory, item, EXT_BY_TYPES[key])
                        )
                    else:
                        local_conf.append(
                            handle_relative_path(ref_path, base_directory, item)
                        )
                else:
                    # in-line JS file
                    local_conf.append(handle_inline(base_directory, item))
        else:
            out_conf[key] = web_config[key]

    return out_conf


URL_FILE_COUNT = 0


async def handle_url(base_directory, entry, ext):
    global URL_FILE_COUNT
    URL_FILE_COUNT += 1
    file_name = f"{URL_FILE_COUNT}{ext}"
    if "?" in file_name:
        file_name = file_name.split("?")[0]
    async with aiohttp.ClientSession() as session:
        async with session.get(entry) as resp:
            if resp.status == 200:
                content = await resp.read()
                with (base_directory / file_name).open(mode="wb") as f:
                    f.write(content)
    return file_name


def handle_relative_path(ref_path, base_directory, entry):
    src = ref_path / entry
    file_name = src.name
    dst = base_directory / file_name
    shutil.copyfile(src, dst)
    return file_name


def handle_inline(base_directory, entry):
    file_name = entry.get("name")
    content = entry.get("content")
    dst = base_directory / file_name
    with dst.open(mode="w") as f:
        f.write(content)
    return file_name


def to_py_attr(file, item, indent=4):
    entry = item.get("name")
    if isinstance(entry, (list, tuple)):
        py_name, js_name = entry
        file.write(f'\n{" "*indent}("{py_name}", "{js_name}"),')
    else:
        file.write(f'\n{" "*indent}"{entry}",')


def to_py_help(file, class_info):
    indent = 4
    file.write(f'\n{" "*indent}"""')
    main_help = class_info.get("help", "")
    properties = class_info.get("properties", [])
    events = class_info.get("properties", [])

    if len(main_help):
        file.write(f'\n{" "* indent}{main_help}')

    if len(properties):
        file.write(f'\n{" "* indent}Properties\n')
        for prop in properties:
            name = prop.get("name")
            help = prop.get("help", "").replace("\n", " ")
            if isinstance(name, (list, tuple)):
                name = name[0]
            file.write(f'\n{" "* indent}:param {name}: {help}')

    if len(events):
        file.write(f'\n\n{" "* indent}Events\n')
        for prop in events:
            name = prop.get("name")
            help = prop.get("help", "").replace("\n", " ")
            if isinstance(name, (list, tuple)):
                name = name[0]
            file.write(f'\n{" "* indent}:param {name}: {help}')
    file.write(f'\n{" "*indent}"""')



import asyncio
import argparse
from .generator import generate_trame_package


async def main():
    parser = argparse.ArgumentParser(description="Trame component generator")

    parser.add_argument(
        "--config",
        default="trame",
        help="Path to config file that specify the widgets definition",
    )

    parser.add_argument(
        "--output",
        help="Directory where the generated files will be stored",
    )

    args, _ = parser.parse_known_args()

    await generate_trame_package(args.config, args.output)


if __name__ == "__main__":
    asyncio.run(main())



from trame.app import get_server
from trame.ui.vuetify import SinglePageLayout
from trame.widgets import vuetify, vtk as vtk_widgets


class Cone:
    def __init__(self, server_or_name=None):
        self.server = get_server(server_or_name, client_type="vue2")
        self.ui = self._generate_ui()

    @property
    def ctrl(self):
        return self.server.controller

    @property
    def state(self):
        return self.server.state

    @property
    def resolution(self):
        return self.state.resolution

    @resolution.setter
    def resolution(self, v):
        with self.state:
            self.state.resolution = v

    def reset_resolution(self):
        self.resolution = 6

    def _generate_ui(self):
        with SinglePageLayout(self.server) as layout:
            layout.title.set_text("Trame demo")
            with layout.toolbar as toolbar:
                toolbar.dense = True
                vuetify.VSpacer()
                vuetify.VSlider(
                    v_model=("resolution", 6),
                    min=3,
                    max=60,
                    step=1,
                    hide_details=True,
                    style="max-width: 300px;",
                )
                with vuetify.VBtn(icon=True, click=self.reset_resolution):
                    vuetify.VIcon("mdi-lock-reset")
                with vuetify.VBtn(icon=True, click=self.ctrl.view_reset_camera):
                    vuetify.VIcon("mdi-crop-free")

            with layout.content:
                with vuetify.VContainer(fluid=True, classes="pa-0 fill-height"):
                    with vtk_widgets.VtkView() as view:
                        self.ctrl.view_reset_camera = view.reset_camera
                        with vtk_widgets.VtkGeometryRepresentation():
                            vtk_widgets.VtkAlgorithm(
                                vtk_class="vtkConeSource", state=("{ resolution }",)
                            )

            return layout


def main(**kwargs):
    cone = Cone()
    cone.server.start(**kwargs)


if __name__ == "__main__":
    main()
